use io;

/* Read instance data. The input files follow the "Solomon" format*/
function input() {
    usage = "Usage: localsolver tdcvrptw.lsp "
            + "inFileName=inputFile [solFileName=outputFile] [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;

    readInputCvrptw();

    computeDistanceMatrices();
}

/* Declare the optimization model */
function model() {
    customersSequences[k in 0...nbTrucks] <- list(nbCustomers);

    // All customers must be visited by exactly one truck
    constraint partition[k in 0...nbTrucks](customersSequences[k]);

    for [k in 0...nbTrucks] {
        local sequence <- customersSequences[k];
        local c <- count(sequence);

        // A truck is used if it visits at least one customer
        truckUsed[k] <- c > 0;

        // The quantity needed in each route must not exceed the truck capacity
        routeQuantity[k] <- sum(sequence, j => demands[j]);
        constraint routeQuantity[k] <= truckCapacity;

        // End of each visit according to the traffic
        endTime[k] <- array(0...c, (i, prev) => max(earliestStart[sequence[i]],
            i == 0 ? travelTimeWarehouse[sequence[0]][timeToMatrixIdx[0]] :
            prev + travelTime[sequence[i-1]][sequence[i]][timeToMatrixIdx[round(prev)]])
            + serviceTime[sequence[i]]);

        // Arriving home after max horizon
        homeLateness[k] <- truckUsed[k] ? max(0, endTime[k][c-1] +
            travelTimeWarehouse[sequence[c-1]][timeToMatrixIdx[round(endTime[k][c-1])]] -
            maxHorizon) : 0;

        // Distance traveled by truck k
        routeDistances[k] <- sum(1...c,
                i => distanceMatrix[sequence[i-1]][sequence[i]])
                + (truckUsed[k] ?
                (distanceDepot[sequence[0]] + distanceDepot[sequence[c-1]]) :
                0);

        // Completing visit after latest end
        lateness[k] <- homeLateness[k] + sum(0...c,
                i => max(0, endTime[k][i] - latestEnd[sequence[i]]));
    }

    // Total lateness, must be 0 for a solution to be valid
    totalLateness <- sum[k in 0...nbTrucks](lateness[k]);

    // Total number of trucks used
    nbTrucksUsed <- sum[k in 0...nbTrucks](truckUsed[k]);

    // Total distance traveled (convention in Solomon's instances is to round to 2 decimals)
    totalDistance <- round(100 * sum[k in 0...nbTrucks](routeDistances[k])) / 100;

    // Objective: minimize the lateness, then the number of trucks used, then the distance traveled
    minimize totalLateness;
    minimize nbTrucksUsed;
    minimize totalDistance;
}


/* Parameterize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

/* Write the solution in a file with the following format:
 * - number of trucks used and total distance
 * - for each truck the customers visited (omitting the start/end at the depot) */
function output() {
    if (solFileName == nil) return;
    local outfile = io.openWrite(solFileName);

    outfile.println(nbTrucksUsed.value, " ", totalDistance.value);
    for [k in 0...nbTrucks] {
        if (truckUsed[k].value != 1) continue;
        for [customer in customersSequences[k].value] {
            outfile.print(customer + 1, " ");
        }
        outfile.println();
    }
}

function readInputCvrptw() {
    local inFile = io.openRead(inFileName);
    skipLines(inFile, 4);

    // Truck related data
    nbTrucks = inFile.readInt();
    truckCapacity = inFile.readInt();

    skipLines(inFile, 3);

    // Depot data
    local line = inFile.readln().split();
    depotIndex = line[0].toInt();
    depotX = line[1].toInt();
    depotY = line[2].toInt();
    maxHorizon = line[5].toInt();

    // Customers data
    i = 0;
    while (!inFile.eof()) {
        inLine = inFile.readln();
        line = inLine.split();
        if (count(line) == 0) break;
        if (count(line) != 7) throw "Wrong file format";
        customerIndex[i] = line[0].toInt();
        customerX[i] = line[1].toInt();
        customerY[i] = line[2].toInt();
        demands[i] = line[3].toInt();
        serviceTime[i] = line[6].toInt();
        earliestStart[i] = line[4].toInt();
        // in input files due date is meant as latest start time
        latestEnd[i] = line[5].toInt() + serviceTime[i];
        i = i + 1;
    }
    nbCustomers = i;

    inFile.close();
}

function skipLines(inFile, nbLines) {
    for [i in 0...nbLines] inFile.readln();
}

// Compute the distance matrices
function computeDistanceMatrices() {
    shortDistanceTravelTimeProfile = {1.00, 2.50, 1.75, 2.50, 1.00};
    mediumDistanceTravelTimeProfile = {1.00, 2.00, 1.50, 2.00, 1.00};
    longDistanceTravelTimeProfile = {1.00, 1.60, 1.10, 1.60, 1.00};
    travelTimeProfileMatrix = {
        shortDistanceTravelTimeProfile,
        mediumDistanceTravelTimeProfile,
        longDistanceTravelTimeProfile
    };
    distanceLevels = {10, 25};
    timeIntervalSteps = {0.0, 0.2, 0.4, 0.6, 0.8, 1.0};
    nbTimeIntervals = count(timeIntervalSteps)-1;
    nbDistanceLevels = count(distanceLevels);

    for [i in 0...nbCustomers] {
        distanceMatrix[i][i] = 0;

        for [k in 0...nbTimeIntervals] {
            travelTime[i][i][k] = 0;
        }

        for [j in i+1...nbCustomers] {
            local localDistance = computeDist(i, j);
            distanceMatrix[j][i] = localDistance;
            distanceMatrix[i][j] = localDistance;

            local profileIdx = getProfile(localDistance);
            for [k in 0...nbTimeIntervals] {
                local localTravelTime = travelTimeProfileMatrix[profileIdx][k] * localDistance;
                travelTime[i][j][k] = localTravelTime;
                travelTime[j][i][k] = localTravelTime;
            }
        }
    }

    for [i in 0...nbTimeIntervals] {
        local timeStepStart = round(timeIntervalSteps[i] * maxHorizon);
        local timeStepEnd = round(timeIntervalSteps[i+1] * maxHorizon);
        for [j in timeStepStart...timeStepEnd+1] {
            timeToMatrixIdx[j] = i;
        }
    }

    for [i in 0...nbCustomers] {
        local localDistance = computeDepotDist(i);
        distanceDepot[i] = localDistance;

        local profileIdx = getProfile(localDistance);
        for [j in 0...nbTimeIntervals] {
            local localTravelTimeWareHouse = travelTimeProfileMatrix[profileIdx][j] * localDistance;
            travelTimeWarehouse[i][j] = localTravelTimeWareHouse;
        }
    }
}

function computeDist(i, j) {
    local x1 = customerX[i];
    local x2 = customerX[j];
    local y1 = customerY[i];
    local y2 = customerY[j];
    return computeDistance(x1, x2, y1, y2);
}

function computeDepotDist(i) {
    local x1 = customerX[i];
    local xd = depotX;
    local y1 = customerY[i];
    local yd = depotY;
    return computeDistance(x1, xd, y1, yd);
}

function computeDistance(x1, x2, y1, y2) {
    return sqrt(pow((x1 - x2), 2) + pow((y1 - y2), 2));
}

function getProfile(distance) {
    local idx = 0;
    while (idx < nbDistanceLevels && distance > distanceLevels[idx]) {
        idx = idx + 1;
    }
    return idx;
}